@@ -9,14 +9,37 @@ use crate::bytecode::InternalBytecode;
 use crate::error::{message_bad_eof, message_from_token, message_incorrect_requirement};
 use crate::expr::{Expr, ExprParser};
 use crate::lex::{Token, TokenContext, TokenStream, TokenType, Version};
+use crate::CustomInstruction;
 
 /// The number of gates that are built in to the OpenQASM 2 language.  This is U and CX.
 const N_BUILTIN_GATES: usize = 2;
-/// The number of gates that 'qelib1.inc' defines.  For efficiency, this parser doesn't actually
-/// parse the include file itself, since its contents are fixed by the OpenQASM 2 arXiv paper,
-/// which this parser follows to the letter.  Instead, we just have it hardcoded as a special
-/// import that happens quickly when the relevant 'include' statement is seen.
-const N_QELIB1_GATES: usize = 23;
+/// The "qelib1.inc" special include.  The elements of the tuple are the gate name, the number of
+/// parameters it takes, and the number of qubits it acts on.
+const QELIB1: [(&str, usize, usize); 23] = [
+    ("u3", 3, 1),
+    ("u2", 2, 1),
+    ("u1", 1, 1),
+    ("cx", 0, 2),
+    ("id", 0, 1),
+    ("x", 0, 1),
+    ("y", 0, 1),
+    ("z", 0, 1),
+    ("h", 0, 1),
+    ("s", 0, 1),
+    ("sdg", 0, 1),
+    ("t", 0, 1),
+    ("tdg", 0, 1),
+    ("rx", 1, 1),
+    ("ry", 1, 1),
+    ("rz", 1, 1),
+    ("cz", 0, 2),
+    ("cy", 0, 2),
+    ("ch", 0, 2),
+    ("ccx", 0, 3),
+    ("crz", 1, 2),
+    ("cu1", 1, 2),
+    ("cu3", 3, 2),
+];
 
 /// A symbol in the global symbol table.  Parameters and individual qubits can't be in the global
 /// symbol table, as there is no way for them to be defined.
@@ -34,6 +57,8 @@ enum GlobalSymbol {
         n_params: usize,
         n_qubits: usize,
         index: usize,
+        custom: bool,
+        defined: bool,
     },
 }
 
@@ -115,14 +140,22 @@ pub struct State {
 
 impl State {
     /// Create and initialise a state for the parser.
-    pub fn new(tokens: TokenStream, include_path: Vec<std::path::PathBuf>) -> Self {
+    pub fn new(
+        tokens: TokenStream,
+        include_path: Vec<std::path::PathBuf>,
+        custom_instructions: &[CustomInstruction],
+    ) -> Result<Self, String> {
         let mut state = State {
             tokens: vec![tokens],
             context: TokenContext::new(),
             include_path,
             // For Qiskit-created circuits, all files will have the builtin gates and `qelib1.inc`,
-            // so we allocate with that in mind.
-            symbols: HashMap::with_capacity(N_BUILTIN_GATES + N_QELIB1_GATES),
+            // so we allocate with that in mind.  There may well be overlap between libraries and
+            // custom instructions, but this is small-potatoes allocation and we'd rather not have
+            // to reallocate.
+            symbols: HashMap::with_capacity(
+                N_BUILTIN_GATES + QELIB1.len() + custom_instructions.len(),
+            ),
             gate_symbols: HashMap::new(),
             n_qubits: 0,
             n_clbits: 0,
@@ -130,10 +163,29 @@ impl State {
             n_gates: 0,
             allow_version: true,
         };
+        for inst in custom_instructions {
+            if state
+                .symbols
+                .insert(
+                    inst.name.clone(),
+                    GlobalSymbol::Gate {
+                        n_params: inst.n_params,
+                        n_qubits: inst.n_qubits,
+                        index: state.n_gates,
+                        custom: true,
+                        defined: inst.name == "U" || inst.name == "CX" || inst.builtin,
+                    },
+                )
+                .is_some()
+            {
+                return Err(format!("duplicate custom instruction '{}'", inst.name));
+            }
+            state.n_gates += 1;
+        }
         let dummy_token = Token::dummy();
-        state.define_gate(&dummy_token, "U".into(), 3, 1).unwrap();
-        state.define_gate(&dummy_token, "CX".into(), 0, 2).unwrap();
-        state
+        state.define_gate(&dummy_token, "U".into(), 3, 1)?;
+        state.define_gate(&dummy_token, "CX".into(), 0, 2)?;
+        Ok(state)
     }
 
     /// Get the next token available in the stack of token streams, popping and removing any
@@ -627,7 +679,22 @@ impl State {
                 n_params,
                 n_qubits,
                 index,
-            }) => Ok((*index, *n_params, *n_qubits)),
+                custom,
+                defined,
+            }) => {
+                if *custom && !defined {
+                    Err(message_from_token(
+                        &name_token,
+                        &format!(
+                            "cannot use non-builtin custom instruction '{}' before definition",
+                            name,
+                        ),
+                        self.current_filename(),
+                    ))
+                } else {
+                    Ok((*index, *n_params, *n_qubits))
+                }
+            }
             Some(symbol) => Err(message_from_token(
                 &name_token,
                 &format!(
@@ -1167,31 +1234,26 @@ impl State {
             .int(&self.context);
         self.expect(TokenType::RBracket, "']'", &lbracket_token)?;
         self.expect(TokenType::Semicolon, "';'", &creg_token)?;
-        bc.push(Some(InternalBytecode::DeclareCreg {
-            name: name.clone(),
-            size,
-        }));
-        if self
-            .symbols
-            .insert(
-                name,
-                GlobalSymbol::Creg {
-                    size,
-                    start: self.n_clbits,
-                    index: self.n_cregs,
-                },
-            )
-            .is_some()
-        {
-            return Err(message_from_token(
+        match self.symbols.insert(
+            name.clone(),
+            GlobalSymbol::Creg {
+                size,
+                start: self.n_clbits,
+                index: self.n_cregs,
+            },
+        ) {
+            None | Some(GlobalSymbol::Gate { defined: false, .. }) => {
+                self.n_clbits += size;
+                self.n_cregs += 1;
+                bc.push(Some(InternalBytecode::DeclareCreg { name, size }));
+                Ok(1)
+            }
+            _ => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is already defined", name_token.id(&self.context)),
                 self.current_filename(),
-            ));
+            )),
         }
-        self.n_clbits += size;
-        self.n_cregs += 1;
-        Ok(1)
     }
 
     /// Parse a declaration of a quantum register, emitting the relevant bytecode and adding the
@@ -1207,29 +1269,24 @@ impl State {
             .int(&self.context);
         self.expect(TokenType::RBracket, "']'", &lbracket_token)?;
         self.expect(TokenType::Semicolon, "';'", &qreg_token)?;
-        bc.push(Some(InternalBytecode::DeclareQreg {
-            name: name.clone(),
-            size,
-        }));
-        if self
-            .symbols
-            .insert(
-                name,
-                GlobalSymbol::Qreg {
-                    size,
-                    start: self.n_qubits,
-                },
-            )
-            .is_some()
-        {
-            return Err(message_from_token(
+        match self.symbols.insert(
+            name.clone(),
+            GlobalSymbol::Qreg {
+                size,
+                start: self.n_qubits,
+            },
+        ) {
+            None | Some(GlobalSymbol::Gate { defined: false, .. }) => {
+                self.n_qubits += size;
+                bc.push(Some(InternalBytecode::DeclareQreg { name, size }));
+                Ok(1)
+            }
+            _ => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is already defined", name_token.id(&self.context)),
                 self.current_filename(),
-            ));
+            )),
         }
-        self.n_qubits += size;
-        Ok(1)
     }
 
     /// Parse an include statement.  This currently only actually handles includes of `qelib1.inc`,
@@ -1244,8 +1301,14 @@ impl State {
         self.expect(TokenType::Semicolon, "';'", &include_token)?;
         let filename = filename_token.filename(&self.context);
         if filename == "qelib1.inc" {
-            self.include_qelib1(&include_token)?;
-            bc.push(Some(InternalBytecode::SpecialInclude { name: filename }));
+            self.symbols.reserve(QELIB1.len());
+            let mut indices = Vec::with_capacity(QELIB1.len());
+            for (i, (name, n_params, n_qubits)) in QELIB1.iter().enumerate() {
+                if self.define_gate(&include_token, name.to_string(), *n_params, *n_qubits)? {
+                    indices.push(i);
+                }
+            }
+            bc.push(Some(InternalBytecode::SpecialInclude { indices }));
             Ok(1)
         } else {
             let base_filename = std::path::PathBuf::from(&filename);
@@ -1273,36 +1336,6 @@ impl State {
         }
     }
 
-    /// Update the parser state with the built-in version of the `qelib1.inc` include file.  This
-    /// is precisely as the file is described in the arXiv paper.
-    fn include_qelib1(&mut self, include: &Token) -> Result<(), String> {
-        self.symbols.reserve(N_QELIB1_GATES);
-        self.define_gate(include, "u3".to_owned(), 3, 1)?;
-        self.define_gate(include, "u2".to_owned(), 2, 1)?;
-        self.define_gate(include, "u1".to_owned(), 1, 1)?;
-        self.define_gate(include, "cx".to_owned(), 0, 2)?;
-        self.define_gate(include, "id".to_owned(), 0, 1)?;
-        self.define_gate(include, "x".to_owned(), 0, 1)?;
-        self.define_gate(include, "y".to_owned(), 0, 1)?;
-        self.define_gate(include, "z".to_owned(), 0, 1)?;
-        self.define_gate(include, "h".to_owned(), 0, 1)?;
-        self.define_gate(include, "s".to_owned(), 0, 1)?;
-        self.define_gate(include, "sdg".to_owned(), 0, 1)?;
-        self.define_gate(include, "t".to_owned(), 0, 1)?;
-        self.define_gate(include, "tdg".to_owned(), 0, 1)?;
-        self.define_gate(include, "rx".to_owned(), 1, 1)?;
-        self.define_gate(include, "ry".to_owned(), 1, 1)?;
-        self.define_gate(include, "rz".to_owned(), 1, 1)?;
-        self.define_gate(include, "cz".to_owned(), 0, 2)?;
-        self.define_gate(include, "cy".to_owned(), 0, 2)?;
-        self.define_gate(include, "ch".to_owned(), 0, 2)?;
-        self.define_gate(include, "ccx".to_owned(), 0, 3)?;
-        self.define_gate(include, "crz".to_owned(), 1, 2)?;
-        self.define_gate(include, "cu1".to_owned(), 1, 2)?;
-        self.define_gate(include, "cu3".to_owned(), 3, 2)?;
-        Ok(())
-    }
-
     /// Update the parser state with the definition of a particular gate.  This does not emit any
     /// bytecode because not all gate definitions need something passing to Python.  For example,
     /// the Python parser initialises its state including the built-in gates `U` and `CX`, and
@@ -1313,24 +1346,68 @@ impl State {
         name: String,
         n_params: usize,
         n_qubits: usize,
-    ) -> Result<(), String> {
-        if self.symbols.contains_key(&name) {
-            Err(message_from_token(
+    ) -> Result<bool, String> {
+        match self.symbols.get_mut(&name) {
+            None => {
+                self.symbols.insert(
+                    name,
+                    GlobalSymbol::Gate {
+                        n_params,
+                        n_qubits,
+                        index: self.n_gates,
+                        custom: false,
+                        defined: true,
+                    },
+                );
+                self.n_gates += 1;
+                Ok(true)
+            }
+            Some(GlobalSymbol::Gate {
+                n_params: defined_n_params,
+                n_qubits: defined_n_qubits,
+                custom: true,
+                defined,
+                ..
+            }) => {
+                if n_params != *defined_n_params || n_qubits != *defined_n_qubits {
+                    let plural = |count: usize, singular: &str| {
+                        let mut out = format!("{} {}", count, singular);
+                        if count != 1 {
+                            out.push('s');
+                        }
+                        out
+                    };
+                    let from_custom = format!(
+                        "{} and {}",
+                        plural(*defined_n_params, "parameter"),
+                        plural(*defined_n_qubits, "qubit")
+                    );
+                    let from_program = format!(
+                        "{} and {}",
+                        plural(n_params, "parameter"),
+                        plural(n_qubits, "qubit")
+                    );
+                    Err(message_from_token(
+                        owner,
+                        &format!(
+                            concat!(
+                                "custom instruction '{}' is mismatched with its definition: ",
+                                "OpenQASM program has {}, custom has {}",
+                            ),
+                            name, from_program, from_custom
+                        ),
+                        self.current_filename(),
+                    ))
+                } else {
+                    *defined = true;
+                    Ok(false)
+                }
+            }
+            _ => Err(message_from_token(
                 owner,
                 &format!("name '{}' is already defined", name),
                 self.current_filename(),
-            ))
-        } else {
-            self.symbols.insert(
-                name,
-                GlobalSymbol::Gate {
-                    n_params,
-                    n_qubits,
-                    index: self.n_gates,
-                },
-            );
-            self.n_gates += 1;
-            Ok(())
+            )),
         }
     }
 
